/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic, Ales Zamuda                     *
 *   borko.boskovic@uni-mb.si, ales.zamuda@uni-mb.si                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * \file board.h
 * This file defines the Board as a bitboard and its operations.
 * Board it the current state of game. As players move the pieces,
 * the Board changes. There is only one single Board in program, declared in
 * engine.c. The single Board is created on heap.
 */

#ifndef BOARD_H
#define BOARD_H

#include "move.h"

#define MAX_PLY 500				/**< Maximum number of plys that game could deveop (= 250 full-moves) */
#define MAX_GENERATED_MOVES 150 /**< Maximum number of moves that could be generated in a ply */
#define EVAL_MATE 5000			/**< The evaluation of position that results to mate (position in mate)*/
#define EVAL_DRAW 0
#define MAX_PV 10

/** Represents the full state of game.
 * Holds:
 *	- positions of pieces in member "pieces",
 *	- links to individual types of pieces (WHITE, BLACK, "pawns", "knights", "bishops", "rooks", "queens", "king"),
 *	- free squares, not occupied by any piece ("freeSquares") and its bit-negation ("notFreeSquares"),
 *	- rotated positions of pieces for generating moves with rotated bitboard ("piecesRL90", "piecesRL45", "piecesRR45"),
 *	- Zobrist hash key of the position ("key"),
 *	- history of moves and their Zobrist keys that led to current position ("historyOfMoves", "historyOfKeys"),
 *	- current ply (search recursion depth) investigating ("ply"),
 *	- player (white or black) on the move in the current position ("player") and the opponont player, not on the move ("opponent"),
 *	- the en-passant square of the opponent and its history ("enPassant", "undoEnPassant"),
 *	- availability of four castling squares ("castling") and history information for castling ("undoCastling"),
 *	- whether king is currently in check and recursive history of that ("check", "undoCheck"),
 *	- referential board with types of pieces on the squares ("types"): e.g. a1 contains a white rook.
 *	- number of peaces each player has on board and their material worth ("numberOfPieces", "material"),
 *	- information for prevention of 50 stalled moves rule resulting in a draw ("numberOfStaticMoves"),
 *	- history of which piece was taken down in each move ("undoCapturePiece").
 *
 * Usage:
 * bitboard is used in almost all search operations, as it represents the current game state.
 * When search is parallelized, each thread has it's own bitboard. Board is copied when splitting the search tree.
  */
struct bitboard{
	guint64* pawns;				/**< pointer to pawns pieces for black and white (start=8 bits set for each) */
	guint64* knights;			/**< pointer to knights pieces for black and white (start=2 bits set for each) */
	guint64* bishops;			/**< pointer to bishops pieces for black and white (start=2 bits set for each) */
	guint64* rooks;				/**< pointer to rooks pieces for black and white (start=2 bits set for each) */
	guint64* queens;			/**< pointer to queens pieces for black and white (start=1 bit set for each) */
	guint64* king;				/**< pointer to kings pieces for black and white (start=1 bits set for each) */
	guint64 pieces[14];			/**< bit positions for white, black and separate for all their piece types */
	guint64 freeSquares;		/**< unoccupied squares */
	guint64 notFreeSquares;		/**< occupied squares */
	guint64 piecesRL90;			/**< board with pieces, rotated for 90 deg. left */
	guint64 piecesRL45;			/**< board with pieces, rotated for 45 deg. left */
	guint64 piecesRR45;			/**< board with pieces, rotated for 45 deg. right */
	guint64 key;				/**< Zobrist hash key of the position */
	guint64 historyOfKeys[MAX_PLY]; 	/**< history stack: Zobrist keys of the moves that have led to current position (for detection of three repetitive positions) */
	Move    historyOfMoves[MAX_PLY];	/**< history stack: history of moves since the start position to the current ply position */
	guint   ply;				/**< current search depth (half-moves) */
	guint   player;				/**< player on the move (black or white) */
	guint   opponent;			/**< opponent of the current player (black or white) */
	guint   enPassant;			/**< the en-passant square of the opponent (0-1 bits set) */
	guint   castling;			/**< availability of castling and two castling squares: one square to the left and right of the king's start position for each king (0-6 bits set) */
	guint   check;							/**< boolean, whether king is in check */
	guint   types[64];						/**< type of the figures on each square */
	guint   numberOfPieces[2];				/**< number of pieces each player has on the board */
	gint    material[2];					/**< material worth of each side's pieces (e.g. material(single pawn) = 100) */
	guint   numberOfStaticMoves[MAX_PLY]; 	/**< history stack: number of moves that could lead to 50 moves rule because the game was static */
	guint   undoCapturePiece[MAX_PLY]; 		/**< history stack: state of the square that was occupied by the moved piece (piece could have been captured by doing that) */
	guint   undoCheck[MAX_PLY];				/**< history stack: boolean, if on the last position the king was in check */
	guint   undoCastling[MAX_PLY];			/**< history stack: castling status undo (max. of 6 bits set) */
	guint   undoEnPassant[MAX_PLY]; 		/**< history stack: en passant location in the position of previous ply */
};

/**
 * We chose bitboard representation for the Board.
 */
typedef struct bitboard Board;

struct gmoves{
	gint size;
	gint sorted;
	gint next;
	Move move[MAX_GENERATED_MOVES];
	gint value[MAX_GENERATED_MOVES];
};

typedef struct gmoves Moves;

struct rootMove{
	Move move;
	gint value;
	gint depth;
	gint seldepth;
	guint nps;
	gdouble time;
	guint nodes;
	guint hash;
	Move pv[MAX_PV+1];
};

typedef struct rootMove RootMove;

struct rootMoves{
	guint size;
	guint multipv;
	RootMove move[MAX_GENERATED_MOVES];
};

typedef struct rootMoves RootMoves;


/**
 * Initializes Zobrist keys and attack constants.
 * \param key hash keys of the squares
 * \param player player hash key
 */
void boardInit(gint64 key[64][64], gint64 player);

/**
 * Creates a new board.
 * \return Returns a new, initialized board at start position.
 */
Board* boardNew();

/**
 * Frees the given board.
 * \param board given board
 */
void boardFree(Board* board);

/**
 * Sets given board data to reflect the start position.
 * \param board given board
 */
void boardStartPosition(Board* board);

/**
 * Prints given board with its pieces to standard output.
 * \param board given board
 */
void boardPrint(const Board* board);

guint boardThreeTimeRepeat(const Board* board);
guint board50MovesRule(const Board* board);

/**
 * Returns, if king is in check at given board, no extra calculation is done.
 * \param board given board
 * \return whether king is in check
 */
gboolean boardIsInCheck(const Board* board);

/**
 * Returns current player, no extra calculation is done.
 * \param board board to consider
 * \return current player on the move
 */
guint boardCurrentPlayer(const Board* board);

/**
 * Changes position on board by moving piece(s), removing possible captured pieces and updating history stacks.
 * History stacks are also updated to help undoing the move:
 *	- historyOfMoves,
 *	- undoCapturePiece
 *	- undoCastling,
 *	- undoEnPassant,
 *	- undoCheck,
 *	- historyOfKeys and
 * 	- numberOfStaticMoves.
 * \param board board to consider
 * \param move move to make
 */
gboolean boardMakeMove(Board* board, Move move);

gboolean boardMakeNextMove(Board* board, Moves* moves);

/**
 * Undoes the last made move. Information about last move is taken from history stacks:
 *	- historyOfMoves,
 *	- undoCapturePiece
 *	- undoCastling,
 *	- undoEnPassant,
 *	- undoCheck,
 *	- historyOfKeys and
 * 	- numberOfStaticMoves.
 * \param board board to consider
 */
Move boardUnmakeMove(Board* board);

/**
 * Makes null move by not moving any piece, but updating:
 * - position hash key,
 * - en passant position,
 * - changing player on the move,
 * - storing history and updating and
 * - castling history.
 * \param board board to consider
 */
void boardMakeNullMove(Board* board);

/**
 * Makes undo of the null move by updating:
 * - position hash key,
 * - en passant position,
 * - changing player on the move,
 * - storing history and updating and
 * - castling history.
 * \param board board to consider
 */
void boardUnmakeNullMove(Board* board);		

/**
 * Generates legal moves, that can be done on the given board.
 * \param board board to consider
 * \param moves generated legal moves
 * \param historyHeuristicsMatrix moves frequency matrice, cumulative history
 */
void boardGenerateNonQuiescentMoves(const Board* board, Moves* moves, gint historyHeuristicsMatrix[64][64]);

/**
 * Generates moves that are changing the board drastically:
 *	- capturing of pieces,
 *	- promotion of the pawns and
 *	- risky check positions?
 * \param board board to consider
 * \param moves generated moves list to write into
 */
void boardGenerateQuiescentMoves(const Board* board, Moves* moves);

/**
 * Builds up a board setup from the FEN specification, describing the 8 ranks x 8 files board.
 * Start position fen is: "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -", meaning:
 *	- figures in the top (8th) rank are: black rook, knight, bishop, queen, king, bishop, knight and rook,
 *	- 7th rank contains 8 pawns,
 *	- there are for ranks empty between,
 *	- eight white pawns follow,
 *	- remaining white figures in the bottom.
 *	- "w" - white is on the move,
 *	- "K" - white can still castle on the king (right) side,
 *	- "Q" - white can still castle on the queen (left) side,
 *	- "k" and "q" - same castling situation for the black player and
 *	- "-" - there are no additional moves played after the fen (long notation would follow, e.g. "e2e4")
 * \param board board to consider
 * \param fen character string with fen notation for the position
 */
void boardSetFen(Board* board, gchar fen[]);

/**
 * Applies long notation moves in a string to the given board.
 * \param board given board to consider
 * \param moves long notation moves string to apply
 */
void boardMakeMovesFromString(Board* board, gchar moves[]);

/**
 * Prints single bitboard layer data: one type of (pieces on) board.
 * \param x single bitboard layer
 */
void boardBitPrint(guint64 x);

/**
 * Another function to print bitboard (see boardBitPrint), different (with reference indexes) implementation.
 * \param x single bitboard layer
 */
void boardChessBitPrint(guint64 x);

/**
 * Evaluates the board position.
 * Currently, only material worth (updated during application of moves) is considered.
 * \param board given board to consider
 * \param alpha worst possible evaluation until now
 * \param beta best possible evaluation until now
 * \return estimate of the given board position worth for the current player; bad is negative (-) and  good is positive (+)
 */
gint boardEvaluate(const Board* board, gint alpha, gint beta);

/**
 * Validates, whether all the bitboard layers for representation of the board are consistent.
 * \param board given board (bitboard layers) to check for consistency
 */
void boardValidate(const Board* board);

void boardSortMoves(Moves* moves);

void boardGenerateMoves(Board* board, Moves* moves);

void boardGenerateRootMoves(Board* board, RootMoves* moves);

#endif
