/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/** Creates Zobrist hash key table and initializes its constants.
 * Created are keys for:
 * - all 64 squares,
 * - black player.
 * The squareKey architecture for 64 squares (0..63):
 *  0..1: empty square (this Zobrist key is not mandatory actually,
 *  just here to be consistent with other (e.g. pieceValue[14])
 *  array lengths.
 *  2..13: pieces for black/white per two entries interchanging.
 *  
 * \param key already computed random Zobrist keys of pieces on
 * board
 * \param player black player's Zobrist hash key
 */
inline void boardInitZobristKeys(gint64 key[14][64], gint64 player){
	gint i,j;
	if(key == 0){
		GRand* rand = g_rand_new_with_seed(0); /* good for testing */
		for(i=0; i<14; i++) 
			for(j=0; j<64; j++) 
				squareKey[i][j] = g_rand_int(rand) | ((gint64)g_rand_int(rand) << 32);
		playerKey = g_rand_int(rand) | ((gint64)g_rand_int(rand) << 32) ;
		g_rand_free(rand);
	}
	else{
		for(i=0; i<14; i++) 
			for(j=0; j<64; j++) 
				squareKey[i][j] = key[i][j];
		playerKey = player;
	}
}

/** Pre-computes attack masks for all piece types from all possible squares.
 */
inline void boardInitAttack(){
	gint s, b, q, f, r; /* square, bit, diagonal, file, rank*/
	guint64 mask;
	for (s=0;s<64;s++) { /* pre-compute attack masks for w,b pawns: two side squares */
		mask = 1ull<<s;
		attackPawn[WHITE][s]  = shiftUpLeft(mask)   | shiftUpRight(mask);
		attackPawn[BLACK][s]  = shiftDownLeft(mask) | shiftDownRight(mask);
	}
	for(s=0; s<64; s++){ /** pre-compute 8 knight's attack squares */
		mask = 1ull<<s;
		attackKnight[s] = shiftLeft(shift2Up(mask))    | shiftRight(shift2Up(mask))   |
				  shiftLeft(shift2Down(mask))  | shiftRight(shift2Down(mask)) |
				  shift2Left(shiftUp(mask))    | shift2Right(shiftUp(mask))   |
				  shift2Left(shiftDown(mask))  | shift2Right(shiftDown(mask));
	}
	for (s=0;s<64;s++) {
		for (b=0;b<64;b++) { 
			mask = 0;
			q = s;
			while (file(q)>0 && rank(q)<7) {
				q += 7;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateL45[q])>>shiftL45[s])) break;
			}
			q = s;
			while (file(q)<7 && rank(q)>0) {
				q -= 7;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateL45[q])>>shiftL45[s])) break;
			}
			attackBishopL45[s][b] = mask;
			mask = 0;
			q = s;
			while (file(q)<7 && rank(q)<7) {
				q += 9;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateR45[q])>>shiftR45[s])) break;
			}
			q = s;
			while (file(q)>0 && rank(q)>0) {
				q -= 9;
				mask |= 1ull<<q;
				if (b&(1ull<<(rotateR45[q])>>shiftR45[s])) break;
			}
			attackBishopR45[s][b] = mask;
		}
	}
	for (s=0;s<64;s++) {
		for (b=0;b<64;b++) {
			f = file(s);
			q = s+1;
			while (++f<8) {
				attackRookL00[s][b]|=1ull<<q;
				if ((1<<f)&(b<<1)) break;
				++q;
			}
			f = file(s);
			q = s-1;
			while (--f>=0) {
				attackRookL00[s][b]|= 1ull<<q;
				if ((1<<f)&(b<<1)) break;
				--q;
			}
			r = rank(s);
			q = s+8;
			while (++r<8) {
				attackRookL90[s][b] |= 1ull<<q;
				if ((1<<(7-r))&(b<<1)) break;
				q+=8;
			}
			r = rank(s);
			q = s-8;
			while (--r>=0) {
				attackRookL90[s][b] |= 1ull<<q;
				if ((1<<(7-r))&(b<<1)) break;
				q-=8;
			}
		}
	}
	for (s=0;s<64;s++) {
	 	mask = 1ull<<s;
	 	attackKing[s] = shiftLeft(mask) | shiftRight(mask) |
				shiftUp(mask)   | shiftDown(mask)  |
				shiftUpLeft(mask) | shiftUpRight(mask) |
				shiftDownLeft(mask) | shiftDownRight(mask);
		maskBishop[s] = attackBishopR45[s][0] | attackBishopL45[s][0];
		maskRook[s] = attackRookL00[s][0] | attackRookL90[s][0];
		maskQueen[s] =	maskBishop[s] | maskRook[s];
		maskKnight[s] = attackKnight[s];
	}
	for(s=0; s<64; s++){
		mask = attackKing[s];
		maskKingSafe[s] = mask;/* | shiftUpLeft(mask) | shiftUpRight(mask) |
								shiftDownLeft(mask) | shiftDownRight(mask);*/
	}
}

inline void boardInitMobility(){
	 gint s, b;
	 for(s=0; s<64; s++){
	 	mobilityKnight8[s] = numberOfBits64(attackKnight[s]);
	 	for(b=0; b<64; b++){
			mobilityBishopL45[s][b] = numberOfBits64(attackBishopL45[s][b]);
			mobilityBishopR45[s][b] = numberOfBits64(attackBishopR45[s][b]);
			mobilityRookL00[s][b] = numberOfBits64(attackRookL00[s][b]);
			mobilityRookL90[s][b] = numberOfBits64(attackRookL90[s][b]);
		}
	}
}

/** Initializes the game by initializing the board.
 * Zobrist hash key table and player's Zobrist hash key are
 * initialized at beginning and attack mask table is created.
 */
void boardInit(gint64 key[14][64], gint64 player){
	boardInitZobristKeys(key, player);
	boardInitAttack();
	boardInitMobility();
	isInit = TRUE;
}
